module marq_mod

  use dal
  use errorHandling
  
  implicit none
  
contains
  
  subroutine dcvsrt(covar,ncvm,ma,lista,mfit)
!c
!cc  repacks the covarience matrix for mqmlmn
!c
!c************************ fform version 1.0 ************  1-mar-88 14:40
!c
!ca  author : press et al.      date:
!cu  update : j. bloch          date:
!cu  update : sls               date:  1-mar-1988 14:39
!c
!ct  status: not tested
!c
!c   general description 
!cg  dcvsrt is the real*8 version of routine covsrt found in "numerical
!cg  recipes" by press et al. see p 515 for details. the only difference
!cg  is that array covar is real*8 instead of single precision.  given
!cg  the covarience matrix covar of a fit mfit of ma total parameters,
!cg  and there ordering lista, repack the covarience matrix to the true
!cg  order of the parameters.  elements associated with fixed parameters
!cg  will be zero.  ncvm is the physical dimension of covar
!c
!c   call_var.          type i/o description 
!cp  ma                  i4  i   total number of parameters
!cp  mfit                i4  i   number of active parameters
!cp  ncvm                i4      dimension of covar array
!cp  covar               r8  i/o covarience matrix
!cp  lista               i4  i   position vector
!c
!c   include_block name          description 
!c
!c   routines_called    type     description 
!c
!c   extensions/system calls     description 
!cx 
!c
!c***********************************************************************
!c!
!c   variables   meaning

    implicit none

    integer :: i, j
    integer, intent(in) :: ma, mfit, ncvm
    integer, intent(inout), dimension(mfit) :: lista
    real(kind=double), intent(inout), dimension(ncvm,ncvm) :: covar
    real(kind=double) :: swap
!c
!c  zero all elements below the diagonal
!c
    do 12 j=1,ma-1
       do 11 i=j+1,ma
          covar(i,j) = 0.
11     enddo
12  enddo
!c
!c  repack off-diagonal elements of fit into correct locations
!c  below the diagonal
!c
    do 14 i=1,mfit-1
       do 13 j=i+1,mfit
          if(lista(j) .gt. lista(i)) then
             covar(lista(j),lista(i)) = covar(i,j)
          else
             covar(lista(i),lista(j)) = covar(i,j)
          endif
13     enddo
14  enddo
!c
!c  temporarily store origial elements in top row, and zero the diagonal
!c
    swap = covar(1,1)
!c
    do 15 j=1,ma
       covar(1,j) = covar(j,j)
       covar(j,j) = 0.
15  enddo
!c
    covar(lista(1),lista(1)) = swap
!c
!c  now sort the elements into proper order on diagonal
!c
    do 16 j=2,mfit
       covar(lista(j),lista(j)) = covar(1,j)
16  enddo
!c
!c  finally, fill in above the diagonal by symmetry
!c
    do 18 j=2,ma
       do 17 i=1,j-1
          covar(i,j) = covar(j,i)
17     enddo
18  enddo
!c
    return
  end subroutine dcvsrt
!c
!c
!c
  subroutine dgausj(a,n,np,b,m,mp,ierr)
!c
!cc  numerical routines linear equation solution routine
!c
!c************************ fform version 1.0 ************  1-mar-88 15:03
!c
!ca  author : press et al.      date:
!cu  update : j. bloch          date:  6-sep-1987
!cu  update : sls               date:  1-mar-1988 14:42
!c
!ct  status: not tested
!c
!c   general description 
!cg  linear equation solution by gauss-jordan elimination.  a is an
!cg  input matrix of n x n elements, stored in an array of physical
!cg  dimensions np x np.  b is an input matrix of n x m containing the
!cg  m right-hand side vectors, stored in an array of physical
!c!g  dimensions np x mp.  on output, a is replaced by its matrix
!cg  inverse, and b is replaced by the corresponding set of solution
!cg  vectors.
!cg  modified by j. bloch to be a double precision matrix version of the
!cg  subroutine gaussj in the book numerical recipes by press et al.
!c
!c   call_var.          type i/o description 
!cp  a                   r8  i/o updated real*8 matrix which will be
!cp                              replaced by it's inverse.
!cp  n                   i4  i   actual order of matrix a
!cp  np                  i4      actual dimenstions of array a.
!cp  b                   r8  i/o updated array containing m right hand
!cp                                  side vectors. upon return, b
!cp                                  contains the solution vectors.
!cp  m                   i4  i   number of right hand side vectors.
!cp  mp                  i4  i   actual second dimension size of b.
!cp  ierr                i4  i/o = 0 no errors
!cp                              = 3 singular matrix
!c
!c   include_block name          description 
!c
!c   routines_called    type     description 
!c
!c   extensions/system calls     description 
!cx 
!c
!c***********************************************************************
!c
!c   variables   meaning
!c
    implicit none
!c
!c       parameter nmax=200
!c
    integer :: i, icol, ierr, irow, j, k, l, ll
    integer, intent(in) :: m, mp, n, np
    integer, dimension(200) :: ipiv, indxr, indxc
    real(kind=double), intent(inout) :: b(np,mp), a(np,np)
    real(kind=double) :: big, dum, pivinv
!c
    ierr = 0
    do 11 j=1,n
       ipiv(j) = 0
11  enddo
!c
!c  start the main loop over the columns to be reduced
!c
    i = 0
    do while ((i .lt. n) .and. (ierr .eq. 0))
       i = i + 1
       big = 0.d0
!c
!c  start the outer loop of the search for the pivot element
!c
       j = 0
       do while ((j .lt. n) .and. (ierr .eq. 0))
          j = j + 1
          if(ipiv(j) .ne. 1) then
             k = 0
             do while ((k .lt. n) .and. (ierr .eq. 0))
                k = k + 1
                if (ipiv(k) .eq. 0) then
                   if (abs(a(j,k)) .ge. big) then
                      big = abs(a(j,k))
!                      big = dabs(a(j,k)) ???????
                      irow = j
                      icol = k
                   endif
                else if (ipiv(k) .gt. 1) then
                   ierr = 3
                   call warning("SingularMatrix",&
                        "DGAUSS Singular Matrix. Point 1")
                endif
             enddo
          endif
       enddo
       if(ierr .eq. 0) then
          ipiv(icol) = ipiv(icol) + 1
!c
!c  we now have the pivot element, so we interchange rows, if needed,
!c  to put the pivot element on the diagonal.  the columns are not
!c  physically interchanged, only relabeled: indx(i), the column of the
!c  ith pivot element, is the ith column that is reduced, while indxr(i)
!c  is the row in which that pivot element was origially located.  if
!c  indxr(i) =/ indxc(i) there is an implied column interchange.  with
!c  this form of bookkeeping, the solution b's will end up in the correct
!c  order, and the inverse matrix will be scrambled by columns.
!c
          if (irow .ne. icol) then
             do 14 l=1,n
                dum = a(irow,l)
                a(irow,l) = a(icol,l)
                a(icol,l) = dum
14           enddo
!c
             do 15 l=1,m
                dum = b(irow,l)
                b(irow,l) = b(icol,l)
                b(icol,l) = dum
15           enddo
          endif
!c
!c  we are now ready to divide the pivot row by the pivot element
!c  located at irow and icol
!c
          indxr(i) = irow
          indxc(i) = icol
          if (a(icol,icol) .eq. 0.) then
             ierr = 3
             call warning("SingularMatrix",&
                  "DGAUSS Singular Matrix. Point 2")
          else
             pivinv = 1.d0/a(icol,icol)
             a(icol,icol) = 1.d0
             do 16 l=1,n
                a(icol,l) = a(icol,l)*pivinv
16           enddo
!c
             do 17 l=1,m
                b(icol,l) = b(icol,l)*pivinv
17           enddo
!c
!c  next we reduce the rows, except for the pivot row
!c
             do 21 ll=1,n
                if(ll .ne. icol)then
                   dum = a(ll,icol)
                   a(ll,icol) = 0.
                   do 18 l=1,n
                      a(ll,l) = a(ll,l) - a(icol,l)*dum
18                 enddo
                   do 19 l=1,m
                      b(ll,l) = b(ll,l) - b(icol,l)*dum
19                 enddo
                endif
21           enddo
          endif
       endif
22  enddo
!c
!c  this is the end of the main loop over the columns of the reduction.
!c  now unscramble the solution in view of the column interchanges.
!c  we do this by interchanging pairs of columns in the order that the
!c  permutation was built up.
!c
    if(ierr .eq. 0) then
       do 24 l=n,1,-1
          if(indxr(l).ne.indxc(l))then
             do 23 k=1,n
                dum = a(k,indxr(l))
                a(k,indxr(l)) = a(k,indxc(l))
                a(k,indxc(l)) = dum
23           enddo
          endif
24     enddo
    endif
    return
  end subroutine dgausj
!c
!c
!c
  subroutine gauss_fit(tempx,tempy,tempu,n,np,ap,ierr)

    integer :: ftype = 2
    integer, intent(inout) :: n, np
    integer :: i, ierr, iflag, iup, iuplim, lista(10), maxitr, nppo, npts
    real(kind=double), intent(inout) :: &
         tempx(n), tempy(n), tempu(n), ap(10)
    real(kind=double) :: a(10), alamda, almlim, chisq, &
         deltch, ochisq, olamda
    real(kind=double) :: alpha(10,10), covar(10,10)
    character(len=6) :: pshort(10) = (/'norm  ', 'width ', 'offset', &
         '      ', '      ', '      ', '      ', '      ', '      ', '      '/)
    if(np .ne. 3) then ! does it *ever* get here?
       call error("nParams",&
            "Wrong number of parameters for a gaussian fit")
    endif
    do i=1,np
       lista(i) = i
       a(i) = ap(i)
    enddo
    nppo = np
!c
!c  *********************************************************************
!c  *                                                                   *
!c  *                       main fit begins here                        *
!c  *                                                                   *
!c  *********************************************************************
!c
    maxitr = 100 ! vary this???
    almlim = 1.e3
    iuplim = 10
    npts = 10
!c
!c        call report(-1,' initial model value settings:',a,fix)
!c
    alamda = -1.0
    ochisq = 0.0
    iup = 0
    i = 1
    iflag = 0
!c
    do while ((iflag .eq. 0) .and. (i .lt. maxitr))
       i = i + 1
       olamda = alamda
       call mqm (ftype,tempx,tempy,tempu,n,a,npts,lista,nppo,&
            covar,alpha,npts,chisq,alamda,ierr)
!c
       deltch = abs(ochisq - chisq)             
!c
!c  check for various termination conditions
!c
       if((alamda .lt. olamda) .and. (deltch .lt. 0.1)) iflag = 1
       if(iflag .eq. 0) then
          if(alamda .gt. olamda) iup = iup + 1
          if(alamda .lt. olamda) iup = 0
          if(iup .gt. iuplim) iflag = 1
          if(alamda .gt. almlim) iup = iup + 1
       endif
    enddo
    write(errstr,*)'np =',np,' n=',n,' m.l.e. = ',chisq

!    write(errstr,1010) ' np =',np,' n =',n,' m.l.e. =',chisq
!1010 format(/,1x,a,i2,a,i4,a,g12.4,a,g14.7)
    call message(AppMsg,Verbose)
!c
!c  finish off the fit to get the covarience matrix
!c
    alamda = 0.0
    call mqm (ftype,tempx,tempy,tempu,n,a,npts,lista,nppo,&
         covar,alpha,npts,chisq,alamda,ierr)
!c
!c  write out final values and one sigma uncertainties
!c
    write(errstr,*) &
         "GAUSS_FIT: Final parameter values and the formal standard errors"
    call message(AppMsg,Verbose)
!c
    do i=1,nppo
       ap(i) = a(i)
       write(errstr,2100)pshort(lista(i)),a(lista(i)),&
            sqrt(abs(covar(lista(i),lista(i))))
!            dsqrt(dabs(covar(lista(i),lista(i))))
2100   format(1x,a15,f13.5,' +/-',f11.5)
       call message(AppMsg,Verbose)
    enddo
!c

    return
  end subroutine gauss_fit
!c
!c
!c
  subroutine gfuncs (tempx,aloc,ymod,dyda,dy2da2,ma)
!c
!cc  calculates gaussian function distribution and partial derivatives
!c
!c************************ fform version 1.0 ************ 21-apr-88 08:24
!c
!ca  author : sls               date: 17-august-2005 15:00  
!c
!ct  status: not tested
!c
!c!   general description 
!c!g  pfuncs is the model value subroutine for gauss_fit for use with the
!cg  fitting subroutine mqm.for. 
!c
!c   call_var.          type i/o description 
!cp  tempx               r4  i/? x value.
!cp  aloc                r4  i/? array of model values.
!cp                              a(1) = normalization
!c!p                              a(2) = width
!cp                              a(3) = peak point
!cp  ymodel              r4      value of the function.
!cp  dyda                r4  i/o array of first partial derivatives
!cp                              w.r.t. each parameter. (only the
!cp                              parameters actually being fit are
!cp                              computed.)
!cp  dy2da2              r4  i/o array of second partial derivatives
!cp                              w.r.t. the parameters.  (only those
!cp                              parameters being fit will be computed.)
!cp  lista               i4  i   list of parameters being fitted.
!cp  mfit                i4  i   number of parameters being fit.
!cp  ma                  i4  i   number of parameters in model.
!cp                              (should be <= 10)
!c
!c   include_block name          description          
!c
!c!   routines_called    type     description 
!c
!c   extensions/system calls     description 
!c
!c***********************************************************************
!c
!    save
!c
    integer, intent(in) :: ma
    integer :: i
    real(kind=double), intent(inout) :: dy2da2(10,10), dyda(10), aloc(ma)
    real(kind=double), intent(in) :: tempx
    real(kind=double) :: a(10), chan, g
    real(kind=double), intent(inout) :: ymod

    do i=1,10
       a(i) = aloc(i)
    enddo
    chan = tempx
!c
!c  compute the model value
!c
    g   = exp(-a(2)*(chan - a(3))*(chan - a(3)))
    ymod = a(1)*g
!        ymodel = sngl(ymod)
!c
!c  compute values on intervals to prepare for computation
!c  of numerical derivatives.
!c
    dyda(1) = g
    dyda(2) = -a(1)*(chan - a(3))*(chan - a(3))*g
    dyda(3) = 2.0*a(1)*a(2)*(chan - a(3))*g
!c
    dy2da2(1,1) = 0.0
    dy2da2(1,2) = -(chan - a(3))*(chan - a(3))*g
    dy2da2(1,3) = 2.0*a(2)*(chan - a(3))*g
    dy2da2(2,1) = -(chan - a(3))*(chan - a(3))*g
    dy2da2(2,2) = a(1)*(chan - a(3))*(chan - a(3))*&
         (chan - a(3))*(chan - a(3))*g
    dy2da2(2,3) = 2.0*a(1)*(chan - a(3))*&
         (1.0 -  a(2)*(chan - a(3))*(chan - a(3)))*g
    dy2da2(3,1) = 2.0*a(2)*(chan - a(3))*g
    dy2da2(3,2) = 2.0*a(1)*(chan - a(3))*&
         (1.0 -  a(2)*(chan - a(3))*(chan - a(3)))*g
    dy2da2(3,3) = 2.0*a(1)*a(2)*&
         (2.0*a(2)*(chan - a(3))*(chan - a(3)) - 1.0)*g
    return
  end subroutine gfuncs
!c
!c
!c
  subroutine mqm (ftype,tempx,tempy,tempu,ndata,a,ma,&
       lista,mfit,covar,alpha,nca,chisq,alamda,ierr)
!c
!cc  maximum likelyhood fitting routine
!c
!c************************ fform version 1.0 ************  2-mar-88 09:16
!c
!ca  author : j. bloch          date:  4-sep-1987
!cu  update : j. bloch          date:  6-sep-1987
!cu  update : sls               date:  2-mar-1988 09:25  
!c
!ct  status: not tested
!c
!c   general description 
!cg  mqmlmn was derived from mrqmin in the book "numerical recipes" by
!cg  press et al. it finds the best parameter fit using the marquardt
!cg  method applied to the poisson maximum likelihood estimator, as
!cg  described by cash in the 1979 march 15 astrophysical journal. the
!cg  sig array has been removed because it is not needed.
!c
!c   call_var.          type i/o description 
!cp  ftype               i4      = 1 polynomial function (pfuncs)
!cp                              = 2 gaussian function (gfuncs)
!cp  ierr                i4  i/o = 0 no error
!cp                              = 3
!cp  tempx               r4  ?/? independent variable data values.
!cp  tempy               r4  ?/? dependent variable data values.
!cp  tempu               r4  ?/? dependent variable uncertainty values.
!cp  ndata               i4      number of data points to fit.
!cp  a                   r4  i/o model parameter values.
!cp  ma                  i4  i/? number of model parameter values.
!cp  lista               i4  i/o list of parameters
!cp                                  in array a actually being varied.
!cp  mfit                i4  i/? number of parameters actually being fit.
!cp  covar               r8  i/o covariance matrix and work
!cp                                  array used durint the fit.
!cp  alpha               r8  i/o matrix alpha used in the marquard method
!cp                                  (see press et al.,pp. 521-528)
!cp  nca                 i4  i/? actual dimensions of alpha and covar.
!cp  chisq               r4  i/o output maximum likelihood estimator.
!cp  alamda              r4  i/o updated real iteration control variable.
!cp                              must be set to  < 0.0 on first call
!cp                              of mqmlmn. if a step fails, almada grows
!cp                              by a factor of 10. on final call, set
!cp                              almada = 0.0 so curvature matrix is 
!cp                              returned.
!p  pfuncs              r4  ?/? input name of model function, with call
!cp                                  sequence:
!cp                                  call pfuncs(x1,a,ymod,dyda,dy2da2,ma)
!cp                                      with:
!cp                                      x1 = independent variable at
!cp                                          which to evaluate the model.
!cp                                      a(ma) = array of model params.
!cp                                      ymod = output model value at x1.
!c!p                                      dyda(ma) = vector of first
!cp                                           derivatives of the model
!cp                                           w.r.t. the parameters.
!cp                                      dy2da2(20,20)= matrix of partial
!cp                                           second derivatives of the
!cp                                           model w.r.t the parameters.
!cp                                      lista = list of parameters being 
!cp                                           fitted.
!cp                                      mfit  = number of parameters being 
!cp                                           fit.
!cp                                      ma = input integer number of
!cp                                           parameters in the model.
!c
!c   include_block name          description 
!c
!c   routines_called    type     description 
!cr  dcvsrt              sr      repacks the covarience matrix for mqm!
!cr  dgausj              sr      numerical routines linear equation 
!cr                              solution routine
!cr  mrq_cb              sr      computes matrix elements for mqm
!c
!c   extensions/system calls     description 
!cx 
!c
!c***********************************************************************
!c
!c   variables   meaning
!c
    implicit none
!c
    integer, intent(inout) :: ma, ndata, nca, mfit
    integer, intent(inout) :: lista(ma)
    integer :: ftype, ierr, ierr1, ihit, j, k, kk

    real(kind=double), intent(inout) :: alamda, chisq
    real(kind=double), intent(inout) :: a(ma), alpha(nca,nca), &
         tempx(ndata), tempy(ndata), tempu(ndata), covar(nca,nca)
    real(kind=double) :: atry(10), beta(10), ochisq
    real(kind=double) :: da(10)
    ierr = 0
    kk = 1
!c
!c  do some initializations
!c
    if(alamda .lt. 0.) then
       kk = mfit + 1
!c
!c  check to see if lista contains a proper permutation of the coeffs.
!c
       j = 0
       do while ((j .lt. ma) .and. (ierr .eq. 0))
          j = j + 1
          ihit = 0
          do 11 k=1,mfit
             if(lista(k) .eq. j) ihit = ihit + 1
11        enddo
          if(ihit .eq. 0) then
             lista(kk) = j
             kk = kk + 1
          else if (ihit .gt. 1) then
             ierr = 3
             call warning("ImpPerm",&
                  "MQM: Improper Permutation in LISTA array. Point 1")  
          endif
       enddo
!c
       if(kk .ne. (ma+1)) then
          ierr = 3
          call warning("ImpPerm",&
               "MQM: Improper Permutation in LISTA array. Point 2")  
       endif
!c
       if(ierr .eq. 0) then
          alamda = 0.001
          call mrq (ftype,tempx,tempy,tempu,ndata,a,ma,&
               lista,mfit,alpha,beta,nca,chisq,ierr1)
          if(ierr1 .eq. 3) then
             ierr = 3
             write(errstr,*) &
                  "MQM: Bad return from MRQ. 1."
             call message(AppMsg,Noisy)
          else
             ochisq = chisq
             do 13 j=1,ma
                atry(j) = a(j)
13           enddo
          endif
       endif
    endif
!c
!c  alter linearized fitting matrix by augmenting diagonal elements
!c
    if(ierr .eq. 0) then
       do 15 j=1,mfit
          do 14 k=1,mfit
             covar(j,k) = alpha(j,k)
14        enddo
          covar(j,j) = alpha(j,j)*(1.0d0 + alamda)
          da(j) = beta(j)
15     enddo
!c
!c  get the matrix solution
!c
       call dgausj(covar,mfit,nca,da,1,1,ierr1)
       if(ierr1 .eq. 3) then
          write(errstr,*) &
               "MQM: Bad return from DGAUSJ. 1."
          call message(AppMsg,Noisy)
          ierr = 3
       else
!c
!c  if converged evaluate covarience matrix with alamda = 0
!c
          if(alamda .eq. 0.) then
             call dcvsrt(covar,nca,ma,lista,mfit)
          else
!c
!c  check to see if the trial succeeded
!c
             do 16 j=1,mfit
!?????                atry(lista(j)) = a(lista(j)) + sngl(da(j))
                atry(lista(j)) = a(lista(j)) + da(j)
16           enddo

             call mrq (ftype,tempx,tempy,tempu,ndata,&
                  atry,ma,lista,mfit,covar,da,nca,chisq,ierr1)
             if(ierr1 .eq. 3) then
                ierr = 3
                write(errstr,*) &
                     "MQM: Bad return from MRQ. 2."
                call message(AppMsg,Noisy)
             else
                if(chisq .lt. ochisq) then
!c
!c  success, accept the new solution
!c
                   alamda = 0.1*alamda
                   ochisq = chisq

                   do 18 j=1,mfit
                      do 17 k=1,mfit
                         alpha(j,k) = covar(j,k)
17                    enddo
                      beta(j) = da(j)
                      a(lista(j)) = atry(lista(j))
18                 enddo
                else
!c
!c  failure, increase alamda and return
!c
                   alamda = 10.*alamda
                   chisq = ochisq
                endif
             endif
          endif
       endif
    endif
    return
  end subroutine mqm
!c
!c
!c
  subroutine mrq (ftype,tempx,tempy,tempu,ndata,a,ma,&
       lista,mfit,alpha,beta,nalp,chisq,ierr)
!c
!cc  computes matrix elements for mqm
!c
!c************************ fform version 1.0 ************ dd-mmm-yy hh:mm
!c
!ca  author : j. bloch          date:  4-sep-1987
!c!u  update : j. bloch          date:  6-sep-1987
!c!u  update : sls               date:  2-mar-1988 09:23  
!c
!ct  status: not tested
!c
!c   general description:
!cg  mrq was derived from mrqcof in the book "numerical recipes"
!cg  by press et al. it computes the matrix elements needed for
!cg  the marquardt method applied to the poisson maximum likelihood
!cg  estimator, as described by cash in the 1979 march 15
!cg  astrophysical journal. the sig array has been removed because
!cg  it is not needed.
!c
!c   call_var.          type i/o description 
!cp  ftype               i4      = 1 polynomial function (pfuncs)
!cp                              = 2 gaussian function (gfuncs)
!cp  ierr                i4    o = 0 no error
!cp                              = 3 illegal model value (model < 0.)
!cp  tempx               r4  ?/? array of independent var. data values.
!cp  tempy               r4  ?/? array of dependent variable data values.
!cp  tempu               r4  ?/? array of dependent variable uncertainty values.
!cp  ndata               i4      number of data points to fit.
!cp  a                   r4  i/o array of model parameter values.
!cp  ma                  i4  i/? number of model parameter values.
!cp  lista               i4  i/o array list of parameters
!cp                                  in array a actually being varied.
!!cp  mfit                i4  i/? number of parameters actually being fit.
!cp  covar               r8  i/o covariance matrix and work
!cp                                  array used durint the fit.
!cp  alpha               r8  i/o matrix alpha used in the marquard 
!cp                              method.  (see press et al., pp. 521-528)
!cp  beta                        vector beta used in the marquardt 
!cp                              method.  (see press et al. pp. 521-528)
!cp  nalp                        actual dimensions of alpha and beta.
!cp  chisq               r4  i/o output maximum likelihood estimator.
!c
!c   include_block_name          description
!ci  r$common:cgenl.cmn          general parameter common block
!c
!c   routines_called    type     description
!cr  hflag               r       output flag handling routine
!cr  wrflag              sr      writes to flag output stream
!cr  pfuncs              sr      input name of model function, with call
!cr                                  sequence:
!cr                              call pfuncs(x1,a,ymod,dyda,dy2da2,lista,
!cr                                          mfit,ma)
!cr                                  with:
!cr                                    x1 = independent variable at
!cr                                        which to evaluate the model.
!cr                                    a(ma) = array of model parameters.
!cr                                    ymod = output model value at x1.
!cr                                    dyda(ma) = vector of first
!c!r                                         derivatives of the model
!cr                                         w.r.t. the parameters.
!cr                                    dy2da2(20,20) = matrix of partial
!cr                                         second derivatives of the
!cr                                         model w.r.t the parameters.
!cr                                    lista(mfit) = same as above.
!cr                                    mfit = same as above.
!cr                                    ma = input integer number of
!cr                                         parameters in the model.
!c
!c   extensions/system calls     description
!cx  
!c
!c***********************************************************************
!c
!c   variables   meaning
!c
    implicit none
!c
!c       parameter mmax=10
!c
    integer :: ftype, i, ierr, j, k, kk
    integer, intent(in) :: ma, mfit, nalp, ndata
    integer, intent(inout) :: lista(mfit)
    real(kind=double), intent(in) :: tempx(ndata), tempy(ndata), tempu(ndata)
    real(kind=double), intent(inout) :: a(ma), chisq, &
         alpha(nalp,nalp), beta(ma) 
    real(kind=double) :: dy, dyda(10), d2yda2(10,10), sig2i, wt, ymod
    ierr = 0
    kk = 0
!c
!c  initialize (symmetric) alpha and beta
!c
    do 12 j=1,mfit
       do 11 k=1,j
          alpha(j,k) = 0.
11     enddo
       beta(j) = 0.
12  enddo

    chisq = 0.0
!c
!c  summation loop over all data
!c
    i = 0
    do while ((i .lt. ndata) .and. (ierr .eq. 0))
       i = i + 1
       if(ftype .eq. 1) then
          !c                call pfuncs(tempx(i),a,ymod,dyda,
          !c     +                  d2yda2,lista,mfit,ma)
       elseif(ftype .eq. 2) then
          call gfuncs(tempx(i),a,ymod,dyda,d2yda2,ma)
       else
!          call error("illegalModel","MRQ: Illegal model type.")
       endif
       if(ymod .lt. -1.0d200) then
          write(errstr,*) "MRQ: Illegal model value =", ymod
          call message(AppMsg,Verbose)
          write(errstr,*) "MRQ:           tempx(i) =", tempx(i)
          call message(AppMsg,Verbose)
          write(errstr,*) "MRQ:           tempy(i) =", tempy(i)
          call message(AppMsg,Verbose)
          write(errstr,*) "MRQ:           tempu(i) =", tempu(i)
          call message(AppMsg,Verbose)
          write(errstr,*) "MRQ:                  a =",&
               (a(kk),kk=1,ma)
          call message(AppMsg,Verbose)
          ierr = 3
       else
          sig2i = 1./(tempu(i)*tempu(i))
          dy = tempy(i) - ymod
         
          do 14 j=1,mfit
             wt = sig2i*dyda(lista(j))
             do 13 k=1,j
                alpha(j,k) = alpha(j,k) + wt*dyda(lista(k))
13           enddo
             beta(j) = beta(j) + dy*wt
14        enddo
          chisq = chisq + dy*dy*sig2i
       endif
    enddo
!c
    if(ierr .eq. 0) then
!c
!c  fill in the symmetric side
!c
       do j=2,mfit
          do k=1,j-1
             alpha(k,j) = alpha(j,k)
          enddo
       enddo
    endif
    return
  end subroutine mrq
end module marq_mod
