!+
! Module: proton_scale_mod
!
! Authors:  Steven L. Snowden - Goddard Space Flight Center (NASA)
!
! Description: Extended Source Analysis Software
!
! Comments:
!
! $Id: proton_scale_mod.f90,v 1.2 2009/02/20 13:27:08 aibarra Exp $
!-
!
!  SAS Version 12.0
!  Version 3.1: 15 March 2012
!
!  Returns the average SP flux and BACKSCAL keyword for a region
!
!  proton-scale uses a mask image created by mos-spectra to find
!  the average SP flare intensity for that region, and returns the
!  BACKSCAL keyword from the spectrum for that region.
!
! Usage:
!   data directory
!   mode: 1 -> single region
!         2 -> multiple regions
!   mode = 1
!       MOS1: 1, MOS2: 2, pn: 3
!       region image file name
!       region spectrum file name
!       Example: 
!             proton-scale /path/qpb 1 1
!                  region-image.fits region-spec.fits
!   mode = 2
!       filename with region information
!       Example:  proton-scale /path/qpb 2 spregion.txt

module proton_scale_mod

  use dal
  use param
  use errorHandling
  use subsn_mod

  implicit none

  private

  public proton_scale

contains

  !---------------------------------------------------------------------
  subroutine proton_scale
  !---------------------------------------------------------------------
        character(len=1) :: xmmd(132)
        character(len=60) :: maskfile, specfile
        character(len=80) :: spfile
        character(len=132) :: xmmdir
        integer :: i, ios, lun, mode, n, num
        integer :: err = 0
        integer :: status = 0
        real :: back, backl(100), c, cl(100)

        equivalence (xmmdir,xmmd)

        xmmdir = stringParameter("caldb")
        mode = intParameter("mode")
        num = intParameter("detector")
        maskfile = stringParameter("maskfile")
        specfile = stringParameter("specfile")
        spfile = stringParameter("spfile")

        print *,'xmmd: ',xmmd
        print *, 'xmmdir: ', xmmdir

        if(mode .eq. 1) then
            print *, 'nam: ', num
            print *, 'maskfile: ', maskfile
            print *, 'specfile: ', specfile

            call spscale(xmmdir,xmmd,num,maskfile, &
                    specfile,back,c)

            print 2001, back
 2001       format(1x,'Area: ',f8.3)
            print *, 'Ave Flux: ', c
            print *, 'Scaled Ave Flux: ', 1.0e5*c*back

        elseif(mode .eq. 2) then
            print *, 'spfile: ', spfile

!  Get an unused lun for the spectrum

            call ftgiou(lun,status)
            if((status .ne. 0) .and. (err .eq. 0)) then
                err = 1
                print *, 'PROTON-SCALE STATUS FTGIOU =', status
                print *, 'POSITION 01'
            endif

            open(unit=lun,form='formatted',status='old',file=spfile)
            ios = 0
            i = 1
            do while (ios .eq. 0)
                read(lun,2010,iostat=ios) maskfile
                read(lun,2010,iostat=ios) specfile
 2010           format(a60)
                if(ios .eq. 0) then
!                    print *, num, maskfile, specfile

                    call spscale(xmmdir,xmmd,num,maskfile, &
                             specfile,backl(i),cl(i))
                    i = i + 1
                endif
            enddo

            n = i - 1
            print *, '  '
            print *, '         Solid   Soft Proton'
            print *, '         Angle     Scaling'
            print *, '  '
            do i=1,n
                print 2000, i, backl(i), &
                        backl(i)*cl(i)/(backl(1)*cl(1))
 2000           format(1x,i5,2f10.4)
            enddo
        endif

        RETURN

  end subroutine proton_scale

end module proton_scale_mod
